import sys
from struct import *
import subprocess

TABLE_ID_OBJECT_ID			= 0x2
TABLE_ID_SCHEMA				= 0x1
TABLE_ID_OBJECT_ID_DUPLICATE		= 0x4
TABLE_ID_CONTAINER			= 0xb
TABLE_ID_CONTAINER_DUPLICATE		= 0xc
TABLE_ID_SCHEMA_DUPLICATE		= 0x6

TABLE_ID_UPCASE 			= 0x7
TABLE_ID_UPCASE_DUPLICATE 		= 0x8
TABLE_ID_LOG_INFORMATION 		= 0x9
TABLE_ID_LOG_INFORMATION_DUPLICATE 	= 0xa
TABLE_ID_VOLUME_INFORMATION 		= 0x500
TABLE_ID_VOLUME_INFORMATION_DUPLICATE 	= 0x501
TABLE_ID_REPARSE_INDEX 			= 0x540
TABLE_ID_REPARSE_INDEX_DUPLICATE 	= 0x541

SIZE_PAGEHEADER 			= 0x50

object_id_table_sistertables = {
	TABLE_ID_UPCASE 			: TABLE_ID_UPCASE_DUPLICATE,
	TABLE_ID_UPCASE_DUPLICATE 		: TABLE_ID_UPCASE,
	TABLE_ID_LOG_INFORMATION 		: TABLE_ID_LOG_INFORMATION_DUPLICATE,
	TABLE_ID_LOG_INFORMATION_DUPLICATE 	: TABLE_ID_LOG_INFORMATION,
	TABLE_ID_VOLUME_INFORMATION 		: TABLE_ID_VOLUME_INFORMATION_DUPLICATE,
	TABLE_ID_VOLUME_INFORMATION_DUPLICATE 	: TABLE_ID_VOLUME_INFORMATION,
	TABLE_ID_REPARSE_INDEX 			: TABLE_ID_REPARSE_INDEX_DUPLICATE,
	TABLE_ID_REPARSE_INDEX_DUPLICATE 	: TABLE_ID_REPARSE_INDEX
}

checkpoint_sistertables = {
	TABLE_ID_OBJECT_ID 			: TABLE_ID_OBJECT_ID_DUPLICATE,
	TABLE_ID_OBJECT_ID_DUPLICATE 		: TABLE_ID_OBJECT_ID,
	TABLE_ID_SCHEMA 			: TABLE_ID_SCHEMA_DUPLICATE,
	TABLE_ID_SCHEMA_DUPLICATE 		: TABLE_ID_SCHEMA,
	TABLE_ID_CONTAINER 			: TABLE_ID_CONTAINER_DUPLICATE,
	TABLE_ID_CONTAINER_DUPLICATE 		: TABLE_ID_CONTAINER
}

tableid_no_translation_needed = [11,12,34]

class Index_Entry:
	def __init__(self, key_index_entry, entry_position, entry_length):
		self.key_index_entry = key_index_entry
		self.entry_position = entry_position
		self.entry_length = entry_length


class Page_Reference:
	def __init__(self, reference_position, input):
		input.seek(reference_position)
		self.lcn0 = unpack('<q', input.read(8))[0]
		self.lcn1 = unpack('<q', input.read(8))[0]
		self.lcn2 = unpack('<q', input.read(8))[0]
		self.lcn3 = unpack('<q', input.read(8))[0]
		input.seek(reference_position + 0x22)
		self.checksumtype = unpack('<b', input.read(1))[0]
		input.seek(reference_position + 0x28)
		self.checksum_position = reference_position + 0x28
		if self.checksumtype == 2:
			self.checksum = unpack('<q', input.read(8))[0]
			self.checksumtype = "CRC64"
			self.checksumlength = 8
		else:
			self.checksum = unpack('<i', input.read(4))[0]
			self.checksumtype = "CRC32"
			self.checksumlength = 4


	def translatelcn(self):
		self.lcn0_translated = translate_LCN(self.lcn0)
		self.lcn1_translated = translate_LCN(self.lcn1)
		self.lcn2_translated = translate_LCN(self.lcn2)
		self.lcn3_translated = translate_LCN(self.lcn3)
		self.calculate_lcn_positions()
		if self.lcn0_translated + 1 != self.lcn1_translated:
			raise Exception("Page is not continous {}".format(self.lcn0_translated))
		if self.lcn1_translated + 1 != self.lcn2_translated:
			raise Exception("Page is not continous {}".format(self.lcn0_translated))
		if self.lcn2_translated + 1 != self.lcn3_translated:
			raise Exception("Page is not continous {}".format(self.lcn0_translated))


	def translateWithoutContainerMapping(self):
		self.lcn0_translated = self.lcn0
		self.lcn1_translated = self.lcn1
		self.lcn2_translated = self.lcn2
		self.lcn3_translated = self.lcn3
		self.calculate_lcn_positions()


	def calculate_lcn_positions(self):
		self.lcn0_position = self.lcn0_translated * cluster_size + bootsector
		self.lcn1_position = self.lcn1_translated * cluster_size + bootsector
		self.lcn2_position = self.lcn2_translated * cluster_size + bootsector
		self.lcn3_position = self.lcn3_translated * cluster_size + bootsector


	def fetch_tableid(self, input):
		input.seek(self.lcn0_position + 0x48)
		self.tableid = unpack('<q', input.read(8))[0]


class MSB_Node:
        def __init__(self, node_position, input):
                self.node_position = node_position
                input.seek(node_position + 0x20)

                self.lcn0 = unpack('<q', input.read(8))[0]
                self.lcn1 = unpack('<q', input.read(8))[0]
                self.lcn2 = unpack('<q', input.read(8))[0]
                self.lcn3 = unpack('<q', input.read(8))[0]

                input.seek(node_position + 0x48)
                self.tableid = unpack('<Q', input.read(8))[0]
                if self.tableid not in tableid_no_translation_needed:
                        self.lcn0_translated = translate_LCN(self.lcn0)
                        self.lcn1_translated = translate_LCN(self.lcn1)
                        self.lcn2_translated = translate_LCN(self.lcn2)
                        self.lcn3_translated = translate_LCN(self.lcn3)
                else:
                        self.lcn0_translated = self.lcn0
                        self.lcn1_translated = self.lcn1
                        self.lcn2_translated = self.lcn2
                        self.lcn3_translated = self.lcn3

                self.lcn0_position = self.lcn0_translated * cluster_size + bootsector
                self.lcn1_position = self.lcn1_translated * cluster_size + bootsector
                self.lcn2_position = self.lcn2_translated * cluster_size + bootsector
                self.lcn3_position = self.lcn3_translated * cluster_size + bootsector

                input.seek(node_position + 0x50)
                self.index_root_size = unpack('<I', input.read(4))[0]
                input.seek(node_position + 0x50 + self.index_root_size + 0xd)
                self.flag = unpack('<B', input.read(1))[0]
                self.children = []
                self.page_refs = []
                self.parent_node = -1
                self.node_entries = []


        def add_child(self, child):
                self.children.append(child)


        def add_ref(self, page_ref):
                self.page_refs.append(page_ref)


        def add_entry_array(self, entry_array):
                self.node_entries = entry_array


        def add_parent_node(self, parent_node):
                self.parent_node = parent_node


class Object_ID_Tablepointers:
        def __init__(self, object_id_table_position, input, container_dict):
                self.position = object_id_table_position
                self.index_entry_array = extract_index_entries(self.position, input)
                self.OTHER_DIRECTORY_TABLES_REF = []
                self.tableid_dict = dict()
                input.seek(object_id_table_position + 0x50)
                self.root_size = unpack('<I', input.read(4))[0]
                input.seek(object_id_table_position + 0x50 + self.root_size + 0xd)
                self.obj_id_flag = unpack('<B', input.read(1))[0]
                self.obj_id_table_msb_tree = 0
                self.obj_id_table_msb_tree = build_msb_tree(object_id_table_position, input)
                self.traverse_object_id_table_tree(self.obj_id_table_msb_tree, input, container_dict)


        def set_table_references(self, input, entry_data, container_dict):
                        input.seek(entry_data.entry_position + 0x4)
                        key_offset = unpack('<H', input.read(2))[0]
                        input.seek(entry_data.entry_position + 0xa)
                        input.seek(entry_data.entry_position + 0x18) # get table id from position offset 0x18
                        tableid = unpack('<q', input.read(8))[0]
                        input.seek(entry_data.entry_position + 0xa)
                        pageref = Page_Reference(entry_data.entry_position + key_offset + 0x30, input)
                        pageref.translatelcn()
                        if tableid == 0x7:
                                self.UPCASE_TABLE_REF = pageref
                        elif tableid == 0x8:
                                self.UPCASE_TABLE_DUP_REF = pageref
                        elif tableid == 0x9:
                                self.LOGFILE_INFORMATION_TABLE_REF = pageref
                        elif tableid == 0xa:
                                self.LOGFILE_INFORMATION_TABLE_DUP_REF = pageref
                        elif tableid == 0xd:
                                self.TRASH_STREAM_TABLE_REF = pageref
                        elif tableid == 0x500:
                                self.VOLUME_INFORMATION_TABLE_REF = pageref
                        elif tableid == 0x501:
                                self.VOLUME_INFORMATION_TABLE_DUP_REF = pageref
                        elif tableid == 0x520:
                                self.FILE_SYSTEM_METADATA_TABLE_REF = pageref
                        elif tableid == 0x530:
                                self.SECURITY_DESCRIPTOR_TABLE_REF = pageref
                        elif tableid == 0x540:
                                self.REPARSE_INDEX_TABLE_REF = pageref
                        elif tableid == 0x541:
                                self.REPARSE_INDEX_TABLE_DUP_REF = pageref
                        elif tableid == 0x600:
                                self.ROOT_DIRECTORY_TABLE_REF = pageref
                        else:
                                other_directory_table = pageref
                                self.OTHER_DIRECTORY_TABLES_REF.append(other_directory_table)


        def traverse_object_id_table_tree(self, tree_node, input, container_dict):
                if tree_node.flag == 0 or tree_node.flag == 2:
                        for elem in tree_node.node_entries:
                                self.set_table_references(input, elem, container_dict)
                elif (tree_node.flag == 1 or tree_node.flag == 3) and tree_node.children:
                        for child in tree_node.children:
                                self.traverse_object_id_table_tree(child, input, container_dict)


        def fetch_tableids(self, input):
                self.UPCASE_TABLE_REF.fetch_tableid(input)
                self.tableid_dict[self.UPCASE_TABLE_REF.tableid] = self.UPCASE_TABLE_REF
                self.UPCASE_TABLE_DUP_REF.fetch_tableid(input)
                self.tableid_dict[self.UPCASE_TABLE_DUP_REF.tableid] = self.UPCASE_TABLE_DUP_REF
                self.LOGFILE_INFORMATION_TABLE_REF.fetch_tableid(input)
                self.tableid_dict[self.LOGFILE_INFORMATION_TABLE_REF.tableid] = self.LOGFILE_INFORMATION_TABLE_REF
                self.LOGFILE_INFORMATION_TABLE_DUP_REF.fetch_tableid(input)
                self.tableid_dict[self.LOGFILE_INFORMATION_TABLE_DUP_REF.tableid] = self.LOGFILE_INFORMATION_TABLE_DUP_REF
                self.TRASH_STREAM_TABLE_REF.fetch_tableid(input)
                self.tableid_dict[self.TRASH_STREAM_TABLE_REF.tableid] = self.TRASH_STREAM_TABLE_REF
                self.VOLUME_INFORMATION_TABLE_REF.fetch_tableid(input)
                self.tableid_dict[self.VOLUME_INFORMATION_TABLE_REF.tableid] = self.VOLUME_INFORMATION_TABLE_REF
                self.VOLUME_INFORMATION_TABLE_DUP_REF.fetch_tableid(input)
                self.tableid_dict[self.VOLUME_INFORMATION_TABLE_DUP_REF.tableid] = self.VOLUME_INFORMATION_TABLE_DUP_REF
                self.FILE_SYSTEM_METADATA_TABLE_REF.fetch_tableid(input)
                self.tableid_dict[self.FILE_SYSTEM_METADATA_TABLE_REF.tableid] = self.FILE_SYSTEM_METADATA_TABLE_REF
                self.SECURITY_DESCRIPTOR_TABLE_REF.fetch_tableid(input)
                self.tableid_dict[self.SECURITY_DESCRIPTOR_TABLE_REF.tableid] = self.SECURITY_DESCRIPTOR_TABLE_REF
                self.REPARSE_INDEX_TABLE_REF.fetch_tableid(input)
                self.tableid_dict[self.REPARSE_INDEX_TABLE_REF.tableid] = self.REPARSE_INDEX_TABLE_REF
                self.REPARSE_INDEX_TABLE_DUP_REF.fetch_tableid(input)
                self.tableid_dict[self.REPARSE_INDEX_TABLE_DUP_REF.tableid] = self.REPARSE_INDEX_TABLE_DUP_REF
                self.ROOT_DIRECTORY_TABLE_REF.fetch_tableid(input)
                self.tableid_dict[self.ROOT_DIRECTORY_TABLE_REF.tableid] = self.ROOT_DIRECTORY_TABLE_REF
                for elem in self.OTHER_DIRECTORY_TABLES_REF:
			elem.fetch_tableid(input)
			self.tableid_dict[elem.tableid] = elem


	def check_tableids(self, tableid):
		if tableid in self.tableid_dict:
			return self.tableid_dict[tableid]
		else:
			return -1


class Superblock:
	def __init__(self, superblock_position, input):
		self.POSITION = superblock_position
		input.seek(superblock_position + 0x50)
		self.GUID1 = unpack('<I', input.read(4))[0] #0x50 - 0x54
		self.GUID2 = unpack('<I', input.read(4))[0] #0x54 - 0x58
		self.GUID3 = unpack('<I', input.read(4))[0] #0x58 - 0x5c
		self.GUID4 = unpack('<I', input.read(4))[0] #0x5c - 0x60
		input.seek(superblock_position + 0x68)
		self.VERSION = unpack('<Q', input.read(8))[0] #0x68 - 0x70
		self.OFFSET_CHECKPOINT_REFERENCES = unpack('<I', input.read(4))[0] #0x70 - 0x74
		self.NUMBER_CHECKPOINT_REFERENCES = unpack('<I', input.read(4))[0] #0x74 - 0x78
		self.OFFSET_SELFDESCRIPTOR = unpack('<I', input.read(4))[0] #0x78 - 0x7c
		self.LENGTH_SELFDESCRIPTOR = unpack('<I', input.read(4))[0] #0x7c - 0x80
		input.seek(superblock_position + self.OFFSET_CHECKPOINT_REFERENCES)
		self.CHECKPOINT1 = unpack('<Q', input.read(8))[0] #checkpoint position 1
		self.CHECKPOINT2 = unpack('<Q', input.read(8))[0] #checkpoint position 2
		self.CHECKPOINT1_TRANSLATED = self.CHECKPOINT1 * cluster_size + bootsector
		self.CHECKPOINT2_TRANSLATED = self.CHECKPOINT2 * cluster_size + bootsector


class Bootsector:
	def __init__(self, bootsector_position, input):
		input.seek(bootsector_position)
		self.ASSEMBLY_INSTRUCTION = input.read(3)
		self.FS_NAME = input.read(8)
		self.ZERO = input.read(5)
		self.FSRS_IDENTIFIER = unpack('<I', input.read(4))[0]
		self.SIZE_VBR = unpack('<H', input.read(2))[0]
		self.CHECKSUM = unpack('<H', input.read(2))[0]
		self.SECTOR_COUNT = unpack('<Q', input.read(8))[0]
		self.BYTES_PER_SECTOR = unpack('<I', input.read(4))[0]
		self.SECTORS_PER_CLUSTER = unpack('<I', input.read(4))[0]
		self.MAJOR_VERSION_NUMBER = unpack('<B', input.read(1))[0]
		self.MINOR_VERSION_NUMBER = unpack('<B', input.read(1))[0]
		input.seek(bootsector_position + 0x38)
		self.VOLUME_SERIAL_NUMBER = unpack('<Q', input.read(8))[0]
		self.BYTES_PER_CONTAINER = unpack('<Q', input.read(8))[0]


class Checkpoint_Tablepointers:
	def __init__(self, checkpoint_position, input):
		self.checkpoint_position = checkpoint_position
		self.references_dict = dict()
		input.seek(self.checkpoint_position + 0x58)	
		self.descriptor_offset = unpack('<i', input.read(4))[0]
		self.descriptor_length = unpack('<i', input.read(4))[0]
		for i in range(13):
			input.seek(self.checkpoint_position + 0x94 + i*4)
			pageref = Page_Reference(unpack('<i', input.read(4))[0] + checkpoint_position , input)
			if i == 0:
				self.OBJECT_ID_TABLE_REFERENCE = pageref
				continue
			if i == 1:
				self.MEDIUM_ALLOCATOR_TABLE_REFERENCE = pageref
				continue
			if i == 2:
				self.CONTAINER_ALLOCATOR_TABLE_REFERENCE = pageref
				continue
			if i == 3:
				self.SCHEMA_TABLE_REFERENCE = pageref
				continue
			if i == 4:
				self.PARENT_CHILD_TABLE_REFERENCE = pageref
				continue
			if i == 5:
				self.OBJECT_ID_DUPLICATE_TABLE_REFERENCE = pageref
				continue
			if i == 6:
				self.BLOCK_REFERENCE_COUNT_TABLE = pageref
				continue
			if i == 7:
				self.CONTAINER_TABLE_REFERENCE = pageref
				self.CONTAINER_TABLE_REFERENCE.translateWithoutContainerMapping()
				continue
			if i == 8:
				self.CONTAINER_DUPLICATE_TABLE_REFERENCE = pageref
				self.CONTAINER_DUPLICATE_TABLE_REFERENCE.translateWithoutContainerMapping()
				continue
			if i == 9:
				self.SCHEMA_DUPLICATE_TABLE_REFERENCE = pageref
				continue
			if i == 10:
				self.CONTAINER_INDEX_TABLE_REFERENCE = pageref
				continue
			if i == 11:
				self.INTEGRITY_STATE_TABLE_REFERENCE = pageref
				continue
			if i == 12:
				self.SMALL_ALLOCATOR_TABLE_REFERENCE = pageref
				self.SMALL_ALLOCATOR_TABLE_REFERENCE.translateWithoutContainerMapping()
				continue


	#only use this AFTER the container table was established
	def translate_table_references(self):
		self.OBJECT_ID_TABLE_REFERENCE.translatelcn()
		self.OBJECT_ID_DUPLICATE_TABLE_REFERENCE.translatelcn()
		self.MEDIUM_ALLOCATOR_TABLE_REFERENCE.translatelcn()
		self.CONTAINER_ALLOCATOR_TABLE_REFERENCE.translatelcn()
		self.SCHEMA_TABLE_REFERENCE.translatelcn()
		self.PARENT_CHILD_TABLE_REFERENCE.translatelcn()
		self.BLOCK_REFERENCE_COUNT_TABLE.translatelcn()
		self.SCHEMA_DUPLICATE_TABLE_REFERENCE.translatelcn()
		self.CONTAINER_INDEX_TABLE_REFERENCE.translatelcn()
		self.INTEGRITY_STATE_TABLE_REFERENCE.translatelcn()


	#only use this AFTER the lcns were translated
	def fetch_tableids(self, input):
		self.OBJECT_ID_TABLE_REFERENCE.fetch_tableid(input)
		self.references_dict[self.OBJECT_ID_TABLE_REFERENCE.tableid] = self.OBJECT_ID_TABLE_REFERENCE
		self.OBJECT_ID_DUPLICATE_TABLE_REFERENCE.fetch_tableid(input)
		self.references_dict[self.OBJECT_ID_DUPLICATE_TABLE_REFERENCE.tableid] = self.OBJECT_ID_DUPLICATE_TABLE_REFERENCE
		self.SMALL_ALLOCATOR_TABLE_REFERENCE.fetch_tableid(input)
		self.references_dict[self.SMALL_ALLOCATOR_TABLE_REFERENCE.tableid] = self.SMALL_ALLOCATOR_TABLE_REFERENCE
		self.MEDIUM_ALLOCATOR_TABLE_REFERENCE.fetch_tableid(input)
		self.references_dict[self.MEDIUM_ALLOCATOR_TABLE_REFERENCE.tableid] = self.MEDIUM_ALLOCATOR_TABLE_REFERENCE
		self.CONTAINER_ALLOCATOR_TABLE_REFERENCE.fetch_tableid(input)
		self.references_dict[self.CONTAINER_ALLOCATOR_TABLE_REFERENCE.tableid] = self.CONTAINER_ALLOCATOR_TABLE_REFERENCE
		self.SCHEMA_TABLE_REFERENCE.fetch_tableid(input)
		self.references_dict[self.SCHEMA_TABLE_REFERENCE.tableid] = self.SCHEMA_TABLE_REFERENCE
		self.PARENT_CHILD_TABLE_REFERENCE.fetch_tableid(input)
		self.references_dict[self.PARENT_CHILD_TABLE_REFERENCE.tableid] = self.PARENT_CHILD_TABLE_REFERENCE
		self.BLOCK_REFERENCE_COUNT_TABLE.fetch_tableid(input)
		self.references_dict[self.BLOCK_REFERENCE_COUNT_TABLE.tableid] = self.BLOCK_REFERENCE_COUNT_TABLE
		self.CONTAINER_TABLE_REFERENCE.fetch_tableid(input)
		self.references_dict[self.CONTAINER_TABLE_REFERENCE.tableid] = self.CONTAINER_TABLE_REFERENCE 
		self.CONTAINER_DUPLICATE_TABLE_REFERENCE.fetch_tableid(input)
		self.references_dict[self.CONTAINER_DUPLICATE_TABLE_REFERENCE.tableid] = self.CONTAINER_DUPLICATE_TABLE_REFERENCE
		self.SCHEMA_DUPLICATE_TABLE_REFERENCE.fetch_tableid(input)
		self.references_dict[self.SCHEMA_DUPLICATE_TABLE_REFERENCE.tableid] = self.SCHEMA_DUPLICATE_TABLE_REFERENCE
		self.CONTAINER_INDEX_TABLE_REFERENCE.fetch_tableid(input)
		self.references_dict[self.CONTAINER_INDEX_TABLE_REFERENCE.tableid] = self.CONTAINER_INDEX_TABLE_REFERENCE
		self.INTEGRITY_STATE_TABLE_REFERENCE.fetch_tableid(input)
		self.references_dict[self.INTEGRITY_STATE_TABLE_REFERENCE.tableid] = self.INTEGRITY_STATE_TABLE_REFERENCE


	def check_tableids(self, tableid):
		if tableid in self.references_dict:
			return self.references_dict[tableid]
		else:
			return -1


	def duplicate_tables_mapper(self, tableid):
		self.switch = { 0 : 5,
				5 : 0,
				7 : 8,
				8 : 7,
				3 : 9,
				9 : 3}
		if tableid in self.switch:
			return self.switch[tableid]
		else:
			return -1


def extract_index_entries(page_position, input):
	input.seek(page_position + SIZE_PAGEHEADER)
	size_rootnode = unpack('<i', input.read(4))[0]
	input.seek(page_position + SIZE_PAGEHEADER + size_rootnode + 0x10)
	start_keyindex = unpack('<i', input.read(4))[0]
	key_count = unpack('<i', input.read(4))[0]
	entries = []
	for i in range(key_count):
		input.seek(page_position + SIZE_PAGEHEADER + size_rootnode + start_keyindex + i*4)
		k = unpack('<i', input.read(4))[0] & 0xffff
		input.seek(k + page_position + SIZE_PAGEHEADER + size_rootnode)
		size_entry = unpack('<i', input.read(4))[0]
		entry = Index_Entry(k, k + page_position + SIZE_PAGEHEADER + size_rootnode, size_entry)
		entries.append(entry)
	return entries


def translate_LCN(lcn):
    cluster_offset = lcn % container_size_cluster
    container_index = int(lcn / (2*container_size_cluster))
    translated_lcn = container_table[container_index] + cluster_offset
    return translated_lcn


def build_container_mapping(container_table_lcn, input):
	container_table_position = container_table_lcn * cluster_size + 0x50 + bootsector
	input.seek(container_table_position)
	size_index_root = unpack('<i', vhd_input.read(4))[0]
	position_index_header = container_table_position + size_index_root
	input.seek(position_index_header + 0x10)

	key_entries = []
	key_entries = extract_index_entries(container_table_position - 0x50, input)
	ct = dict()

	for elem in key_entries:
		input.seek(position_index_header + elem.key_index_entry + 0x04)
		key_offset = unpack('<h', input.read(2))[0]
		input.seek(position_index_header + elem.key_index_entry + key_offset)
		key = unpack('<h', input.read(2))[0]
		input.seek(position_index_header + elem.key_index_entry + 0x90 + 0x10)
		mapping = unpack('<q', input.read(8))[0]
		ct[key] = mapping

	return ct


def msb_tree_pages_to_dict(node, pages_dict):
	pages_dict[node.node_position] = (node.node_position, 0x4000, "MSB")
	traverse_msb_tree(node, page_add_callback, pages_dict)


def msb_tree_refs_to_dict(node, refs_dict):
	traverse_msb_tree(node, ref_add_callback, refs_dict)


def traverse_msb_tree(start, callback, callback_param):
	for child in start.children:
		callback(start, child, callback_param)
		traverse_msb_tree(child, callback, callback_param)


def page_add_callback(node1, node2, pages_dict):
	pages_dict[node2.node_position] = (node2.node_position, 0x4000, "MSB")


def ref_add_callback(node1, node2, refs_dict):
	for pager in node1.page_refs:
		refs_dict[pager.checksum_position] = (pager.checksum_position, pager.lcn0_position, pager.checksumtype, pager.checksumlength)


def build_msb_tree(node_position, input, parentNode = -1):
	input.seek(node_position)
	msb = input.read(4)
	if msb != b'MSB+':
		print("incorrect header signature found. Expected MSB+, found:")
		print(msb)
		sys.exit()
	input.seek(node_position + 0x48)
	node_tableid = unpack('<Q', input.read(8))[0]
	node = MSB_Node(node_position, input)
	node.add_parent_node(parentNode)
	#check first size of index root (0x8 means its a non root node)
	input.seek(node_position + 0x50)
	size_index_root = unpack('<i', input.read(4))[0]

	input.seek(node_position + 0x50 + size_index_root + 0xd)
	#read index header, acquire flags
	flags = unpack('<b', input.read(1))[0]
	input.seek(node_position + 0x50 + size_index_root + 0x10)

	entries = extract_index_entries(node_position, input)
	node.add_entry_array(entries)

	#TODO: Check how to handle a stream node
	#different path if the node is an inner node or root node
	#continue recursive, build tree if inner and or root node 
	#return node if leaf node is reached
	if flags == 0x1 or flags == 0x3:
		for elem in entries:
			input.seek(elem.entry_position + 0xa)
			offset_to_reference = unpack('<h', input.read(2))[0]
			page_ref = Page_Reference(elem.entry_position + offset_to_reference, input)
			page_ref.translatelcn()
			input.seek(elem.entry_position + offset_to_reference)
			lcn0 = unpack('<q', input.read(8))[0]
			if node_tableid not in tableid_no_translation_needed:
				lcn0 = translate_LCN(lcn0)
			node.add_child(build_msb_tree(lcn0 * 4096 + bootsector, input, node))
			node.add_ref(page_ref)
	#leaf node
	else:
		return node

	return node


def extract_self_refs(page, ref_dict):
	self_ref_position = None
	checksum_position = None
	sum_length = None
	target = None
	self_ref_length = None
	if isinstance(page, Superblock):
		self_ref_position = page.POSITION + page.OFFSET_SELFDESCRIPTOR
		checksum_position = self_ref_position + 0x28 # or 0x20
		sum_length = 4
		target = page.POSITION
		self_ref_length = page.LENGTH_SELFDESCRIPTOR

	if isinstance(page, Checkpoint_Tablepointers):
		self_ref_position = page.checkpoint_position +  page.descriptor_offset
		checksum_position = self_ref_position + 0x28 # or 0x20
		sum_length = 4
		target = page.checkpoint_position
		self_ref_length = page.descriptor_length

	if sum_length != None:
		ref_dict[checksum_position] = (checksum_position, target, "CRC32", sum_length, self_ref_position, self_ref_length)


def translate_addr(ttable, addr):
	for translation in ttable:
		if addr >= translation[0] and addr < translation[0] + translation[2]:
			return addr - translation[0] + translation[1]
	raise Exception("Could not translate {:016x}".format(addr))


def translate_refs(ttable, refs):
	translated = list()
	for ref in refs.values():
		if len(ref) == 6:
			translated.append((translate_addr(ttable, ref[0]), translate_addr(ttable, ref[1]), ref[2], ref[3], translate_addr(ttable, ref[4]), ref[5]))
		else:
			translated.append((translate_addr(ttable, ref[0]), translate_addr(ttable, ref[1]), ref[2], ref[3]))
	return translated


#TODO: remove duplicate code, move parsing of object id table in a function
#TODO: enable script to find the start of the ReFS (via bootsector)!

offset = int(sys.argv[1]) # for example 128
bootsector = offset * 512

image_path = sys.argv[2]

all_checkpoints = sys.argv[3] == "true"


shasum = subprocess.check_output(["shasum", image_path])
shasum = shasum.split()[0]

pages_to_extract = dict()
references = dict()
with open(image_path, "r+b") as vhd_input:

	#-------------------------------------#
	#Calculate constants of the filesystem#
	#-------------------------------------#
	vhd_input.seek(bootsector + 32)
	bytes_per_sector = unpack('<i', vhd_input.read(4))[0]

	bootsector1 = Bootsector(bootsector, vhd_input)

	vhd_input.seek(bootsector + 64)
	bytes_per_container = unpack('<q', vhd_input.read(8))[0]

	vhd_input.seek(bootsector + 36)
	sectors_per_cluster = unpack('<i', vhd_input.read(4))[0]
	
	cluster_size = sectors_per_cluster * bytes_per_sector #usually 4096 = 4KB
	
	first_superblock = 30 * cluster_size + bootsector

	amount_clusters = int(bootsector1.SECTOR_COUNT / bootsector1.SECTORS_PER_CLUSTER)
	second_superblock = (amount_clusters - 3) * cluster_size + bootsector
	third_superblock = (amount_clusters - 2) * cluster_size + bootsector
	superblock1 = Superblock(first_superblock, vhd_input)
	superblock2 = Superblock(second_superblock, vhd_input)
	superblock3 = Superblock(third_superblock, vhd_input)
	pages_to_extract[first_superblock] = (first_superblock, 0x1000, "SUPB")
	pages_to_extract[second_superblock] = (second_superblock, 0x1000, "SUPB")
	pages_to_extract[third_superblock] = (third_superblock, 0x1000, "SUPB")
	extract_self_refs(superblock1, references)
	extract_self_refs(superblock2, references)
	extract_self_refs(superblock3, references)

	vhd_input.seek(first_superblock + 112)

	checkpoint_offset = unpack('<i', vhd_input.read(4))[0]
	vhd_input.seek(first_superblock + checkpoint_offset)

	#--------------------------------------------------------------------------------------------#
	#Seek both checkpoints, determine which one is the most recent one (checkpoint virtual clock)#
	#--------------------------------------------------------------------------------------------#
	checkpoint1_reference = (unpack('<q', vhd_input.read(8))[0]) * cluster_size + bootsector
	checkpoint2_reference = (unpack('<q', vhd_input.read(8))[0]) * cluster_size + bootsector
	vhd_input.seek(checkpoint1_reference + 96)
	checkpoint1_virtual_clock = unpack('<q', vhd_input.read(8))[0]
	vhd_input.seek(checkpoint2_reference + 96)
	checkpoint2_virtual_clock = unpack('<q', vhd_input.read(8))[0]
	recent_checkpoint = checkpoint1_reference if checkpoint1_virtual_clock > checkpoint2_virtual_clock else checkpoint2_reference #decide on recent checkpoint to use
	old_checkpoint = checkpoint1_reference if checkpoint1_virtual_clock < checkpoint2_virtual_clock else checkpoint2_reference

	#--------------------------------------------------------------------------------------------#
	#Obtain pointer to container table from checkpoint - parse it to enable LCN translation      #
	#--------------------------------------------------------------------------------------------#
	current_checkpoint = Checkpoint_Tablepointers(recent_checkpoint, vhd_input)
	extract_self_refs(current_checkpoint, references)
	if all_checkpoints:
		older_checkpoint = Checkpoint_Tablepointers(old_checkpoint, vhd_input)
		extract_self_refs(older_checkpoint, references)
	index_entries = []

	#BUILD CONTAINER TABLE
	container_table = dict()
	container_table = build_container_mapping(current_checkpoint.CONTAINER_TABLE_REFERENCE.lcn0_translated, vhd_input)
	
	container_size_cluster = int(bytes_per_container / cluster_size) #required for LCN translation ==> contsize_c

	current_checkpoint.translate_table_references() #map the tables of the checkpoint
	current_checkpoint.fetch_tableids(vhd_input)
	if all_checkpoints:
		older_checkpoint.translate_table_references()
		older_checkpoint.fetch_tableids(vhd_input)

	pages_to_extract[current_checkpoint.checkpoint_position] = (current_checkpoint.checkpoint_position, 0x1000, "CHKP")
	checkpoints = [current_checkpoint,]
	if all_checkpoints:
		pages_to_extract[older_checkpoint.checkpoint_position] = (older_checkpoint.checkpoint_position, 0x1000, "CHKP")
		checkpoints.append(older_checkpoint)

	for ref in current_checkpoint.references_dict.values():
		references[ref.checksum_position] = (ref.checksum_position, ref.lcn0_position, ref.checksumtype, ref.checksumlength)
		table_msb = build_msb_tree(ref.lcn0_position, vhd_input)
		msb_tree_pages_to_dict(table_msb, pages_to_extract)
		msb_tree_refs_to_dict(table_msb, references)

	if all_checkpoints:
		for ref in older_checkpoint.references_dict.values():
			references[ref.checksum_position] = (ref.checksum_position, ref.lcn0_position, ref.checksumtype, ref.checksumlength)
			table_msb = build_msb_tree(ref.lcn0_position, vhd_input)
			msb_tree_pages_to_dict(table_msb, pages_to_extract)
			msb_tree_refs_to_dict(table_msb, references)


	object_tables = list()
    
	object_table_content = Object_ID_Tablepointers(current_checkpoint.OBJECT_ID_TABLE_REFERENCE.lcn0_position, vhd_input, container_table)
	object_table_content.fetch_tableids(vhd_input)
	object_tables.append(object_table_content)

	for ref in object_table_content.tableid_dict.values():
		references[ref.checksum_position] = (ref.checksum_position, ref.lcn0_position, ref.checksumtype, ref.checksumlength)
		table_msb = build_msb_tree(ref.lcn0_position, vhd_input)
		msb_tree_pages_to_dict(table_msb, pages_to_extract)
		msb_tree_refs_to_dict(table_msb, references)


	object_table_dup_content = Object_ID_Tablepointers(current_checkpoint.OBJECT_ID_DUPLICATE_TABLE_REFERENCE.lcn0_position, vhd_input, container_table)
	object_table_dup_content.fetch_tableids(vhd_input)
	object_tables.append(object_table_dup_content)

	for ref in object_table_dup_content.tableid_dict.values():
		references[ref.checksum_position] = (ref.checksum_position, ref.lcn0_position, ref.checksumtype, ref.checksumlength)
		table_msb = build_msb_tree(ref.lcn0_position, vhd_input)
		msb_tree_pages_to_dict(table_msb, pages_to_extract)
		msb_tree_refs_to_dict(table_msb, references)

	if all_checkpoints:
		object_table_content_old = Object_ID_Tablepointers(older_checkpoint.OBJECT_ID_TABLE_REFERENCE.lcn0_position, vhd_input, container_table)
		object_table_content_old.fetch_tableids(vhd_input)
		object_tables.append(object_table_content_old)

		for ref in object_table_content_old.tableid_dict.values():
			references[ref.checksum_position] = (ref.checksum_position, ref.lcn0_position, ref.checksumtype, ref.checksumlength)
			table_msb = build_msb_tree(ref.lcn0_position, vhd_input)
			msb_tree_pages_to_dict(table_msb, pages_to_extract)
			msb_tree_refs_to_dict(table_msb, references)


		object_table_dup_content_old = Object_ID_Tablepointers(older_checkpoint.OBJECT_ID_DUPLICATE_TABLE_REFERENCE.lcn0_position, vhd_input, container_table)
		object_table_dup_content_old.fetch_tableids(vhd_input)
		object_tables.append(object_table_dup_content_old)

		for ref in object_table_dup_content_old.tableid_dict.values():
			references[ref.checksum_position] = (ref.checksum_position, ref.lcn0_position, ref.checksumtype, ref.checksumlength)
			table_msb = build_msb_tree(ref.lcn0_position, vhd_input)
			msb_tree_pages_to_dict(table_msb, pages_to_extract)
			msb_tree_refs_to_dict(table_msb, references)

	pages_ordered = list(pages_to_extract.values())
	pages_ordered.sort(key=lambda x: x[0])
	ttable = list()
	with open(shasum + "_compressed.img", "wb") as comp_file:
		with open(shasum + "_translation_table.txt", "w") as ttable_file:
			for p in pages_ordered:
				vhd_input.seek(p[0])
				page_content = vhd_input.read(p[1])
				current_offset = comp_file.tell()
				comp_file.write(page_content)
				ttable_file.write("{:016x} - {:016x} - {:016x} - {}\n".format(p[0], current_offset, p[1], p[2]))
				ttable.append((p[0], current_offset, p[1], p[2]))

	for r in references.values():
		if len(r) == 6:
			print("{:016x} - {:016x} - {} - {} - {:016x} - {}".format(r[0], r[1], r[2], r[3], r[4], r[5]))
		else:
			print("{:016x} - {:016x} - {} - {}".format(r[0], r[1], r[2], r[3]))

	translated_refs = translate_refs(ttable, references)
	with open(shasum + "_checksums.txt", "w") as checksums_file:
		for r in translated_refs:
			if len(r) == 6:
				checksums_file.write("{:016x} - {:016x} - {} - {} - {:016x} - {}\n".format(r[0], r[1], r[2], r[3], r[4], r[5]))
			else:
				checksums_file.write("{:016x} - {:016x} - {} - {}\n".format(r[0], r[1], r[2], r[3]))

	with open(shasum + "_dup_pages.txt", "w") as dup_file:
		duplicates = dict()
		for table_id, name in [(TABLE_ID_OBJECT_ID, "OBJECT"), (TABLE_ID_SCHEMA, "SCHEMA"), (TABLE_ID_CONTAINER, "CONTAINER")]:
			sister_id = checkpoint_sistertables[table_id]
			for chkp in checkpoints:
				translated_orig = translate_addr(ttable, chkp.references_dict[table_id].lcn0_position)
				translated_sister = translate_addr(ttable, chkp.references_dict[sister_id].lcn0_position)
				duplicates[translated_orig] = (translated_sister, name)
		for table_id, name in [(TABLE_ID_UPCASE, "UPCASE"), (TABLE_ID_LOG_INFORMATION, "LOG_INFO"), (TABLE_ID_VOLUME_INFORMATION, "VOLUME_INFO"), (TABLE_ID_REPARSE_INDEX, "REPARSE")]:
			sister_id = object_id_table_sistertables[table_id]
			for objt in object_tables:
				translated_orig = translate_addr(ttable, objt.tableid_dict[table_id].lcn0_position)
				translated_sister = translate_addr(ttable, objt.tableid_dict[sister_id].lcn0_position)
				duplicates[translated_orig] = (translated_sister, name)

		for orig, dub in duplicates.items():
			dup_file.write("{:016x} - {:016x} - {}\n".format(orig, dub[0], dub[1]))

